{******************************}
{* Main unit for LANMsgServer *}
{******************************}

unit server;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Menus, StdCtrls, ExtCtrls, lNetComponents, lNet, config;

type
  TArray = array of string;
  {****************}
  { TfLANMsgServer }
  {****************}
  TfLANMsgServer = class(TForm)
    mMenu: TMainMenu;
    mServer, mServerExit: TMenuItem;
    mSettings, mSettingsConfig: TMenuItem;
    ltcpServer: TLTCPComponent;
    Label1, Label2, Label3, Label4, Label5: TLabel;
    edServerIP, edServerPort: TEdit;
    edServerLog: TMemo;
    shServerStatus: TShape;
    lbClients: TListBox;
    btStart: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure mServerExitClick(Sender: TObject);
    procedure mSettingsConfigClick(Sender: TObject);
    procedure btStartClick(Sender: TObject);
    procedure ltcpServerAccept(aSocket: TLSocket);
    {procedure ltcpServerDisconnect(aSocket: TLSocket);}
    procedure ltcpServerError(const msg: string; aSocket: TLSocket);
    procedure ltcpServerReceive(aSocket: TLSocket);
  private
    iServerPort: Word;
    sServerIP, sMessage, sSource, sDestination: string;
    bStart, bServerConnected: Boolean;
  end;

const
  ConfigFile  = 'server.txt';
  DefaultPort = 5600;

var
  fLANMsgServer: TfLANMsgServer;

implementation

{$R *.lfm}

{ Pause application for n seconds }

procedure Wait(Interval: Integer);

// I previewed this procedure for the case where the server would go offline before
// the clients would be aware of it (send of the "bye" message not done in time)
// As it seems, there is no delay needed (at least not on my system)
// Should you experience any problems, add some delay code here...

begin

end;

{ Read server IP and port from text file }

procedure ServerConfigRead(var ServerIP: string; var ServerPort: Word);

var
  Line: string;
  InFile: Text;

begin
  ServerIP := ''; ServerPort := 0;
  Assign(InFile, ConfigFile); Reset(InFile);
  while not EoF(InFile) do begin
    Readln(InFile, Line);
    if Line <> '' then begin
      if (Length(Line) > 4) and (LeftStr(Line, 3) = 'ip=') then
        ServerIP := Trim(Copy(Line, 4, Length(Line)))
      else if (Length(Line) > 6) and (LeftStr(Line, 5) = 'port=') then
        ServerPort := StrToInt(Trim(Copy(Line, 6, Length(Line))));
    end;
  end;
  Close(InFile);
end;

{ Write server IP and port to text file }

procedure ServerConfigWrite(ServerIP: string; ServerPort: Word);

var
  Line: string;
  OutFile: Text;

begin
  Assign(OutFile, ConfigFile); Rewrite(OutFile);
  Line := 'ip=' + ServerIP; Writeln(OutFile, Line);
  Line := 'port=' + IntToStr(ServerPort); Writeln(OutFile, Line);
  Close(OutFile);
end;

{ Send a message (to one specific or all clients)}

procedure SendMessage(Mess, Dest: string);

var
  I: Integer;

begin
  fLANMsgServer.ltcpServer.IterReset; I := 0;
  while fLANMsgServer.ltcpServer.IterNext do begin
    // For all clients in the internal list
    Inc(I);
    if (Dest = 'all') or (Dest = fLANMsgServer.ltcpServer.Socks[I].PeerAddress) then begin
      // Consider send the message if client is meant to receive it
      if fLANMsgServer.ltcpServer.Iterator.Connected then
        // Only send the message if the client is actually connected
        fLANMsgServer.ltcpServer.SendMessage(Mess, fLANMsgServer.ltcpServer.Iterator);
    end;
  end;
end;

{ Update the clients list box and send client list to all (connected) clients }

procedure UpdateAndSendClientList;

var
  MsgText, IP: string;

begin
  fLANMsgServer.lbClients.Items.Clear;
  MsgText := 'list';
  fLANMsgServer.ltcpServer.IterReset;
  while fLANMsgServer.ltcpServer.IterNext do begin
    // For all clients in the internal list
    if fLANMsgServer.ltcpServer.Iterator.Connected then begin
      // If client is connected, update the listbox and add client IP to the "list" message string
      // "list" message string: list <IP1> <IP2> ...
      IP := fLANMsgServer.ltcpServer.Iterator.PeerAddress;
      fLANMsgServer.lbClients.Items.AddText(IP);
      MsgText += ' ' + IP;
    end;
  end;
  // Send the client list to all clients
  SendMessage(MsgText, 'all');
end;

{****************}
{ TfLANMsgServer }
{****************}

{ Application start: Initialization }

procedure TfLANMsgServer.FormCreate(Sender: TObject);

begin
  bStart := True; bServerConnected := False;
end;

{ Form becoming active: At application start, read server configuration file (if it exists) }

procedure TfLANMsgServer.FormActivate(Sender: TObject);

begin
  if bStart then begin
    // Application start: Read server configuration file (if it exists)
    if FileExists(ConfigFile) then begin
      ServerConfigRead(sServerIP, iServerPort);
    end
    else begin
      sServerIP := ''; iServerPort := DefaultPort;
    end;
    edServerIP.Text := sServerIP; edServerPort.Text := IntToStr(iServerPort);
    bStart := False;
  end;
  if sServerIP = '' then
    edServerIP.SetFocus
  else if iServerPort = 0 then
    edServerPort.SetFocus
  else
    btStart.SetFocus;
end;

{ Menu item "Server > Exit": Exit application }

procedure TfLANMsgServer.mServerExitClick(Sender: TObject);

begin
  if bServerConnected then begin
    SendMessage('bye', 'all');
    Wait(5);
    ltcpServer.Disconnect(True);
  end;
  Close;
end;

{ Menu item "Settings > Server configuration": Set server IP and port (and save to text file) }

procedure TfLANMsgServer.mSettingsConfigClick(Sender: TObject);

begin
  if not bServerConnected then begin
    if iServerPort = 0 then
      fConfig.edPort.Text := ''
    else
      fConfig.edPort.Text := IntToStr(iServerPort);
    fConfig.edIP.Text := sServerIP;
    fConfig.ShowModal;
    if fConfig.sButton = 'save' then begin
      // Settings were changed: Apply and save to file
      iServerPort := fConfig.iServerPort;
      sServerIP   := fConfig.sServerIP;
      edServerPort.Text := IntToStr(iServerPort);
      edServerIP.Text := sServerIP;
      ServerConfigWrite(sServerIP, iServerPort);
    end;
  end;
end;

{ Button "Start/Stop": Start resp. stop the server }

procedure TfLANMsgServer.btStartClick(Sender: TObject);

begin
  if btStart.Caption = 'Start' then begin
    // Button "Start": Start the server
    sServerIP := edServerIP.Text; iServerPort := StrToInt(edServerPort.Text);
    if (sServerIP <> '') and (iServerPort <> 0) then begin
      ltcpServer.Host := sServerIP;
      bServerConnected := ltcpServer.Listen(iServerPort);
      if bServerConnected then begin
        shServerStatus.Brush.Color := clLime;
        edServerLog.Lines.AddText('LAN Messenger Server started at IP=' + sServerIP + ', port=' + IntToStr(iServerPort));
        btStart.Caption := 'Stop';
      end;
    end;
  end
  else begin
    // Button "Stop": Stop the server
    if bServerConnected then begin
      SendMessage('bye', 'all');
      Wait(5);
      ltcpServer.Disconnect(True);
      bServerConnected := False;
      shServerStatus.Brush.Color := clRed;
      lbClients.Items.Clear;
      edServerLog.Lines.AddText('LAN Messenger Server stopped');
      btStart.Caption := 'Start';
    end;
  end;
end;

{ Event triggered when a client connects: Send clients list to all clients }

procedure TfLANMsgServer.ltcpServerAccept(aSocket: TLSocket);

begin
  edServerLog.Lines.AddText('LAN Messenger client with IP=' + aSocket.PeerAddress + ' connected');
  UpdateAndSendClientList;
end;

{ Event triggered when a client disconnects }

{procedure TfLANMsgServer.ltcpServerDisconnect(aSocket: TLSocket);

// This feature seems not to work correctly, the event never being triggered!?

begin
  edServerLog.Lines.AddText('LAN Messenger client with IP=' + aSocket.PeerAddress + ' disconnected');
  UpdateAndSendClientList(aSocket.PeerAddress);
end;}

{ Event triggered when there was a communication error (should not happen...) }

procedure TfLANMsgServer.ltcpServerError(const msg: string; aSocket: TLSocket);

begin
  MessageDlg('Network error', 'LANMsgServer returned the error message: ' + msg, mtError, [mbOK], 0);
end;

{ Event triggered when the server receives a message }

procedure TfLANMsgServer.ltcpServerReceive(aSocket: TLSocket);

var
  P: Integer;

begin
  // Get sender IP and message text
  sSource := aSocket.PeerAddress;  aSocket.GetMessage(sMessage);
  if sMessage = 'bye' then begin
    // // Command "bye" = client will go offline: Send clients list to all clients
    aSocket.Disconnect(True);
    edServerLog.Lines.AddText('LAN Messenger client with IP=' + sSource + ' disconnected');
    UpdateAndSendClientList;
  end
  else if LeftStr(sMessage, 3) = 'msg' then begin
    // Command "msg" = server got a message to relay to a client: Display sender, receiver and message text
    // Server message has format: msg <sender-ip> <message-text>
    Delete(sMessage, 1, 4);
    P := Pos(' ', sMessage);
    sDestination := Copy(sMessage, 1, P - 1);
    Delete(sMessage, 1, P);
    sMessage := 'msg ' + sSource + ' ' + sMessage;
    SendMessage(sMessage, sDestination);
    edServerLog.Lines.AddText('Relayed message from ' + sSource + ' to ' + sDestination);
  end;
end;

end.

